#include "cpgl.h"

extern "C" {
#include "obj.h"
#include "varrier_combiner.h"
#include "tracker.h"
}

#include "model.h"
#include "camera.h"
#include <math.h>
#include <time.h>
#include <unistd.h>
#include <vector>


//--------------------------------------------

#ifndef PI
#define PI 3.14159
#endif

#ifndef E_CONSTANT
#define E_CONSTANT 2.7182818284
#endif

#define DEG2RAD( x ) ( (x) / 180.0f * PI )
#define RAD2DEG( x ) ( (x) / PI * 180.0f )

//-----------------------------------
SDL_Surface* gDrawSurface = NULL;
int   width    = 800;
int   height   = 600;
float aspect   = 4.0f / 3.0f;

bool BreakLoop = false;
bool Capture   = false;
bool DisplayDepth = false;


//-----------------------------------
#define MONO_MODE              1
#define GEOWALL_STEREO_MODE    2
#define PERSONAL_VARRIER_MODE  3

int  DisplayMode = MONO_MODE;

#define    MONO_EYE   0
#define    LEFT_EYE   0
#define    RIGHT_EYE  1
int        Eyes  = 1;

vc_display VCDisp;
float      VCDispShiftDelta = 0.0f;

float      LEP[3] = { -.1, 0.0, 0.0 };
float      REP[3] = { .1,  0.0, 0.0 };

//-----------------------------------
// Update Function Pointer and Variables Updated Each Call
float   TotalTime   = 0;
GLfloat LightPos[4] = { 0, 15, 5, 0};
GLfloat LightVel[3] = { 0, 0, 0 };
GLfloat CamVel[3]   = { 0, 0, 0 };
GLfloat CamPitch    = 0;
GLfloat CamYaw      = 0;
float   FOV         = 60.0f;

//----------------------------------
// Model objects
int RoomModel = -1;
int ToriModel = -1;
int GunModel  = -1;

//---------------------------------
// DOF Related declerations
#define FOCAL_DISTANCE  0
#define LENS_DIAMETER   1
#define ZFOCUS          2
#define MAX_COC         3
#define SCENE_RANGE     4
#define FOCAL_RANGE     5

float DOFCoefficients[6];
float DOFDelta[6];
 
GLuint *ColorBlurFBO;
GLuint **ColorBlurColorTex;
GLuint *ColorBlurDepthTex;

GLuint SceneSampleFBO;
GLuint SceneSampleColorTex;
GLuint SceneSampleDepthTex;

GLhandleARB PassOnePrg;
GLhandleARB PassOneVs;
GLhandleARB PassOneFs;

GLhandleARB PassTwoPrg;
GLhandleARB PassTwoVs;
GLhandleARB PassTwoFs;

GLhandleARB PassThreePrg;
GLhandleARB PassThreeVs;
GLhandleARB PassThreeFs;

bool DrawSceneSampleView = true;
bool DrawPassTwoViews    = true;

#define SCENE_SAMPLE_DIM 32
#define TEXW 1024
#define TEXH 1024
#define COLOR_ATTACH 0
#define BLUR_ATTACH  1

#define NEAR_PLANE   0.1f
#define FAR_PLANE    90.0f

//---------------------------------
// App specific functions
void  Update(double dt);
void  LoadModels();
void  SetupFBO();
void  SetupMultiplePassShaders();
void  SetupSceneSampleFilter();
float RunSceneSampleFilter();

#if !defined(__APPLE__) && !defined(_WIN32)

// SHADER FUNCTIONS
PFNGLCREATEPROGRAMOBJECTARBPROC     glCreateProgramObjectARB = NULL;
PFNGLCREATESHADEROBJECTARBPROC      glCreateShaderObjectARB = NULL;
PFNGLSHADERSOURCEARBPROC            glShaderSourceARB = NULL;
PFNGLCOMPILESHADERARBPROC           glCompileShaderARB = NULL;
PFNGLGETOBJECTPARAMETERIVARBPROC    glGetObjectParameterivARB = NULL;
PFNGLATTACHOBJECTARBPROC            glAttachObjectARB = NULL;
PFNGLGETINFOLOGARBPROC              glGetInfoLogARB = NULL;
PFNGLLINKPROGRAMARBPROC             glLinkProgramARB = NULL;
PFNGLUSEPROGRAMOBJECTARBPROC        glUseProgramObjectARB = NULL;
PFNGLGETUNIFORMLOCATIONARBPROC      glGetUniformLocationARB = NULL;
PFNGLUNIFORM1FARBPROC               glUniform1f = NULL;
PFNGLUNIFORM2FARBPROC               glUniform2f = NULL;
PFNGLUNIFORM3FARBPROC               glUniform3f = NULL;
PFNGLUNIFORM1FVARBPROC              glUniform1fv = NULL;
PFNGLUNIFORM2FVARBPROC              glUniform2fv = NULL;
PFNGLUNIFORM3FVARBPROC              glUniform3fv = NULL;
PFNGLUNIFORM1IARBPROC               glUniform1i = NULL;

PFNGLDELETEOBJECTARBPROC            glDeleteObjectARB= NULL;
PFNGLBINDATTRIBLOCATIONARBPROC      glBindAttribLocationARB= NULL;
PFNGLGETATTRIBLOCATIONARBPROC       glGetAttribLocationARB= NULL;
PFNGLVERTEXATTRIB3FARBPROC          glVertexAttrib3fARB= NULL;

// FBO functions
PFNGLDRAWBUFFERSPROC                glDrawBuffers= NULL;
PFNGLGENFRAMEBUFFERSEXTPROC         glGenFramebuffersEXT= NULL;
PFNGLBINDFRAMEBUFFEREXTPROC         glBindFramebufferEXT= NULL;
PFNGLDELETEFRAMEBUFFERSEXTPROC      glDeleteFramebuffersEXT= NULL;
PFNGLFRAMEBUFFERTEXTURE2DEXTPROC    glFramebufferTexture2DEXT= NULL;
PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC  glCheckFramebufferStatusEXT= NULL;

// MULTITEXTURE FUNCTIONS
PFNGLACTIVETEXTUREPROC              glActiveTexture = NULL;

#endif


//----------------------------------------------------------------
//
//----------------------------------------------------------------
int main(int argc, char** argv)
{
    SetupSDL();
    SetupGLProcs();
    init_model_sys();

    // determine display mode

    if( argc > 2 ) 
    {
        sscanf( argv[1], "%d", &width);
        sscanf( argv[2], "%d", &height);
        if( argc > 3 )
        {
            if( !strcmp(argv[3],"-geowall"))
            {
                DisplayMode = GEOWALL_STEREO_MODE;
                Eyes = 2;
                SetupSDL();
                SetupGLProcs();

                VCDisp.viewport_x = 0;
                VCDisp.viewport_y = 0;
                VCDisp.viewport_w = width;
                VCDisp.viewport_h = height;

                VCDisp.screen_BL[0] = -1.0521f;
                VCDisp.screen_BL[1] = -0.8646f;
                VCDisp.screen_BL[2] = -1.3614f;

                VCDisp.screen_TL[0] = -1.0521f;
                VCDisp.screen_TL[1] =  0.4531f;
                VCDisp.screen_TL[2] = -1.3614f;

                VCDisp.screen_BR[0] =  1.0521f;
                VCDisp.screen_BR[1] = -0.8646f;
                VCDisp.screen_BR[2] = -1.3614f;


            }
            else if( !strcmp(argv[3],"-varrier"))
            {
                // setup varrier_combiner code
                SetupSDL();
                SetupGLProcs();

                if( !vc_init(width,height) )
                {
                    printf("Varrier Combiner Initialization Failed!\n");
                    printf("Exiting Program\n");
                    return 1;
                }


                if( !tracker_init(TRACKER_KEY, CONTROL_KEY) ) 
                {
                    printf("Varrier Tracker Initialization Failed\n");
                    printf("Exiting Program\n");
                    return 1;
                }

                DisplayMode = PERSONAL_VARRIER_MODE;
                Eyes = 2;
                VCDisp.quality = .5f;

                // VALUES THAT THE VARRIER GROUP MEASURED

                VCDisp.viewport_x = 0;
                VCDisp.viewport_y = 0;
                VCDisp.viewport_w = width;
                VCDisp.viewport_h = height;

                VCDisp.screen_BL[0] = -1.0521f;
                VCDisp.screen_BL[1] = -0.8646f;
                VCDisp.screen_BL[2] = -1.3614f;

                VCDisp.screen_TL[0] = -1.0521f;
                VCDisp.screen_TL[1] =  0.4531f;
                VCDisp.screen_TL[2] = -1.3614f;

                VCDisp.screen_BR[0] =  1.0521f;
                VCDisp.screen_BR[1] = -0.8646f;
                VCDisp.screen_BR[2] = -1.3614f;

                VCDisp.pitch = 266.670830f;
                VCDisp.angle =  -7.572000f;
                VCDisp.thick =   0.016000f;
                VCDisp.shift =  -0.000070f;
                VCDisp.cycle =   0.812500f;

                position_camera( 0, 4.0, 0);

            }
        }
        else
        {
            SetupSDL();
            SetupGLProcs();
        }
    }


    // scene setup

    init_model_sys();
    LoadModels();
    InitLight();

    orient_camera(DEG2RAD(0), DEG2RAD(-90), DEG2RAD(0));

    // shader and fbo setup

    SetupFBO();
    SetupMultiplePassShaders();
    SetupSceneSampleFilter();

    CheckGLErrors("Checking for errors before inf loop");

    // run main loop

    InfLoop();

    // free model objects
    
    free_all_models();
    
    vc_fini();
    tracker_fini();

    return 0;
}

//----------------------------------------------------------------
// Load Shader Source From File
//----------------------------------------------------------------
char* GetShaderSource(const char* filename)
{
    char* source;
    FILE* fptr = fopen(filename,"rb");
    if( !fptr )
        return NULL;
    fseek(fptr,0,SEEK_END);
    long length = ftell(fptr);
    source = new char[length + 1];
    memset(source,0,length + 1);
    fseek(fptr,0,SEEK_SET);
    fread(source,sizeof(char),length,fptr);
    fclose(fptr);

    return source;
}

//----------------------------------------------------------------
//
//----------------------------------------------------------------
void UnloadShaders(GLhandleARB& vs, GLhandleARB& fs, GLhandleARB& prg)
{
    glDeleteObjectARB(prg);
}

//----------------------------------------------------------------
// Setup shaders
//----------------------------------------------------------------
void SetupShaders(const char* vert, const char* frag, GLhandleARB& prg, 
                  GLhandleARB& vs, GLhandleARB& fs)
{
    printf("======================\n");
    printf("Setting up Shaders:\nVertex Shader %s\nFragment Shader %s\n",
            vert, frag);
    char infolog[1024];
    GLint length;
    // get shader source code
    char* vertsource = GetShaderSource(vert);
    char* fragsource = GetShaderSource(frag);
    if(!vertsource && !fragsource )
    {
        printf("No shader source code!\n");
        return;
    }

    prg = glCreateProgramObjectARB();
    
    // process the vertex shader source
    if( vertsource )
    {    
        vs = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
        glShaderSourceARB(vs,1,(const GLcharARB**)&vertsource,NULL);
        glCompileShaderARB(vs);
       
        // see if it compiled
        GLint compiled;
        glGetObjectParameterivARB(vs,GL_OBJECT_COMPILE_STATUS_ARB,
            &compiled);
        
        if(compiled)
            glAttachObjectARB(prg,vs);
        else
        {
            printf("VERTEX SHADER COMPILE PROBLEM\n");
            glGetInfoLogARB(vs,1024,&length,infolog);
            if( length ) printf("GLSL PROGRAM LOG: %s\n", infolog);
        }
    }
   
    // process the fragment shader source code
    if( fragsource )
    {
        fs = glCreateShaderObjectARB(GL_FRAGMENT_SHADER);
        glShaderSourceARB(fs,1,(const GLcharARB**)&fragsource,NULL);
        glCompileShaderARB(fs);
       
        // see if the fragment shader compiled
        GLint compiled;
        glGetObjectParameterivARB(fs,GL_OBJECT_COMPILE_STATUS_ARB,
            &compiled);
        
        if(compiled)
            glAttachObjectARB(prg,fs);
        else
        {
            printf("FRAG SHADER COMPILE PROBLEM\n");
            glGetInfoLogARB(fs,1024,&length,infolog);
            if( length ) printf("GLSL PROGRAM LOG: %s\n", infolog);
        }
    }

    // bind 6th vertex attribute to attribute variable name "Tangent"
    glBindAttribLocationARB(prg, 6, "Tangent");
    
    glLinkProgramARB(prg);
    
    glGetInfoLogARB(prg,1024,&length,infolog);
    if( length )
    {
        printf("GLSL PROGRAM LOG: %s\n", infolog );
    }
    
    glUseProgramObjectARB(prg);

    printf("----- DONE ------\n");
    if( fragsource ) delete [] fragsource;
    if( vertsource ) delete [] vertsource;

}

//----------------------------------------------------------------
// Draw To Scene Sample Texture for determining Focal Distance!
//----------------------------------------------------------------
void DrawToSceneSampleTex()
{
    glBindTexture(GL_TEXTURE_2D, 0);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, SceneSampleFBO );
    
    glUseProgramObjectARB( PassOnePrg );

    glViewport( 0, 0, SCENE_SAMPLE_DIM, SCENE_SAMPLE_DIM);
    glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective( FOV * .1, aspect , NEAR_PLANE, FAR_PLANE);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glEnable(GL_LIGHTING);
    
    apply_camera_matrix();

    // make sure that the shaders bound to the models don't override the
    // scene sampling shader

    set_model_override_shader(RoomModel,false);
    set_model_override_shader(ToriModel,false);
    set_model_override_shader(GunModel, false);
    
    draw_model( RoomModel, false);
    draw_model( ToriModel, false);

    float camx, camy, camz;
    float camori[16];

    get_camera_position( &camx, &camy, &camz);
    get_camera_orientation( camori );
    
    position_model ( GunModel, -camx, -camy, -camz);
    orient_model   ( GunModel, camori );
    translate_model( GunModel, .2, .1, 0);

    draw_model(GunModel, false);

    unapply_camera_matrix();

    if(Capture)
    {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
        glBindTexture(GL_TEXTURE_2D, SceneSampleColorTex);
        snap("sample_color.png", SCENE_SAMPLE_DIM, SCENE_SAMPLE_DIM,
             TEXBUF_SRC);
        glBindTexture(GL_TEXTURE_2D, SceneSampleDepthTex);
        snap("sample_depth.png", SCENE_SAMPLE_DIM, SCENE_SAMPLE_DIM,
             TEXBUF_SRC);
    }
}

//========================================================================
// MRT Render in Mono for color and depth/blur maps
//========================================================================
void DrawToColorBlurTex_Mono()
{

    glBindTexture(GL_TEXTURE_2D, 0);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, ColorBlurFBO[0] );
    GLenum bufs[] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };
    glDrawBuffers(2,bufs);

    glViewport( 0, 0, width, height);
    glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective( FOV, aspect, NEAR_PLANE, FAR_PLANE);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glEnable(GL_LIGHTING);
    
    apply_camera_matrix();

    glUseProgramObjectARB(PassTwoPrg);
    glUniform1f(glGetUniformLocationARB(PassTwoPrg,"focal_dist"), 
                DOFCoefficients[FOCAL_DISTANCE]);

    set_model_override_shader(RoomModel,false);
    set_model_override_shader(ToriModel,false);
    set_model_override_shader(GunModel, false);

//    glUseProgramObjectARB(0);
    
    draw_model( RoomModel, false);
    draw_model( ToriModel, false);
    draw_model( GunModel,  false);

    unapply_camera_matrix();

    if( Capture )
    {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[0][0] );
        snap("color.png",width, height,TEXBUF_SRC);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurDepthTex[0] );
        snap("depth.png", width, height, TEXBUF_SRC);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[0][1] );
        snap("blur.png", width, height, TEXBUF_SRC);
    }
}

//========================================================================
// Draw to the on-screen framebuffer in Mono Mode
//========================================================================
void DrawToDefaultFB_Mono()
{

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    GLenum bufs[] = { GL_BACK };
    glDrawBuffers( 1, bufs );

    glViewport( 0, 0, width, height);
    glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D( 0, width, height, 0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    
    glDisable(GL_LIGHTING);
    glDisable(GL_CULL_FACE);
    glDisable(GL_DEPTH_TEST);
    
    // Draw the scene and draw the depth & blur values

    float ms, mt;
    ms = float(width) ;
    mt = float(height);

    glUseProgramObjectARB(PassThreePrg);

    glActiveTexture(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[0][1] );    
    glActiveTexture(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[0][0] );

    glBegin(GL_QUADS);
    glTexCoord2f(0,0);   glVertex2f(0,height);
    glTexCoord2f(0,mt);  glVertex2f(0,0);
    glTexCoord2f(ms,mt); glVertex2f(width,0);
    glTexCoord2f(ms,0);  glVertex2f(width,height);
    glEnd();

    glUseProgramObjectARB(0);

    
    // Draw a cross hair
    glActiveTexture(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
    glActiveTexture(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
    
    glLineWidth(3);
    glColor3f(0, .8, 0);
    glBegin(GL_LINE_STRIP);
    {
        glVertex2f( width / 2 - 12, height / 2 - 12);
        glVertex2f( width / 2 - 12, height / 2 + 12);
        glVertex2f( width / 2 + 12, height / 2 + 12);
        glVertex2f( width / 2 + 12, height / 2 - 12);
        glVertex2f( width / 2 - 12, height / 2 - 12);
    }
    glEnd();
    glLineWidth(1);
   
    glColor3f(1,1,1);
    if( DrawSceneSampleView )
    {
        glBindTexture(GL_TEXTURE_2D, SceneSampleColorTex );

        glBegin(GL_QUADS);
        glTexCoord2f(0,   0);    glVertex2f(20,180);
        glTexCoord2f(0,   1.0);  glVertex2f(20,20);
        glTexCoord2f(1.0, 1.0);  glVertex2f(240,20);
        glTexCoord2f(1.0, 0);    glVertex2f(240,180);
        glEnd();
    }

    if( DrawPassTwoViews )
    {
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[0][1] );
        
        glBegin(GL_QUADS);
        glTexCoord2f(0,  0);    glVertex2f(20,380);
        glTexCoord2f(0,  mt);   glVertex2f(20,220);
        glTexCoord2f(ms, mt);   glVertex2f(240,220);
        glTexCoord2f(ms, 0);    glVertex2f(240,380);
        glEnd();
    }

    if( Capture )
    {
        snap("dfb.png",width,height,FRAME_SRC);
    }
    
}

//========================================================================
// MRT to determine color map and depth_blur map in Stereo Mode
//========================================================================
void DrawToColorBlurTex_Stereo()
{

    // *********** LEFT EYE ***************** //

    glBindTexture(GL_TEXTURE_2D, 0);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, ColorBlurFBO[0] );
    GLenum bufs[] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };
    glDrawBuffers(2,bufs);

    glViewport( 0, 0, width, height);
    glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    vc_frustum( &VCDisp, LEP, NEAR_PLANE, FAR_PLANE);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glEnable(GL_LIGHTING);
    
    apply_camera_matrix();

    glUseProgramObjectARB(PassTwoPrg);
    glUniform1f(glGetUniformLocationARB(PassTwoPrg,"focal_dist"), 
                DOFCoefficients[FOCAL_DISTANCE]);

    set_model_override_shader(RoomModel,false);
    set_model_override_shader(ToriModel,false);
    set_model_override_shader(GunModel, false);

    draw_model( RoomModel, false);
    draw_model( ToriModel, false);
    draw_model( GunModel,  false);

    unapply_camera_matrix();

    if( Capture )
    {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[0][0] );
        snap("color_left.png", width, height,TEXBUF_SRC);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurDepthTex[0] );
        snap("depth_left.png", width, height , TEXBUF_SRC);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[0][1] );
        snap("blur_left.png", width, height , TEXBUF_SRC);
    }

    // *********** RIGHT EYE ***************** //

    glBindTexture(GL_TEXTURE_2D, 0);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, ColorBlurFBO[1] );
    glDrawBuffers(2,bufs);

    glViewport( 0, 0, width, height);
    glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    vc_frustum( &VCDisp, REP, NEAR_PLANE, FAR_PLANE);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glEnable(GL_LIGHTING);
    
    apply_camera_matrix();

    glUseProgramObjectARB(PassTwoPrg);
    glUniform1f(glGetUniformLocationARB(PassTwoPrg,"focal_dist"), 
                DOFCoefficients[FOCAL_DISTANCE]);

    set_model_override_shader(RoomModel,false);
    set_model_override_shader(ToriModel,false);
    set_model_override_shader(GunModel, false);

    draw_model( RoomModel, false);
    draw_model( ToriModel, false);
    draw_model( GunModel,  false);

    unapply_camera_matrix();

    if( Capture )
    {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[1][0] );
        snap("color_right.png",width, height,TEXBUF_SRC);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurDepthTex[1] );
        snap("depth_right.png",width, height , TEXBUF_SRC);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[1][1] );
        snap("blur_right.png", width, height, TEXBUF_SRC);
    }
}

//========================================================================
// Drawing to the on-screen framebuffer in Geowall Stereo Mode
//========================================================================
void DrawToDefaultFB_Geowall()
{
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    GLenum bufs[] = {GL_BACK};
    glDrawBuffers(1, bufs);

    glClearColor(1, 1,0,1);
    glViewport(0,0,width * 2, height);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D( 0, width * 2, height, 0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glDisable(GL_LIGHTING);
    glDisable(GL_CULL_FACE);
    glDisable(GL_DEPTH_TEST);

    float ms, mt;
    ms = float(width) ;
    mt = float(height);

    // DRAW TWO QUADS WITH DOF EFFECT APPLIED

    glUseProgramObjectARB(PassThreePrg);

    glActiveTexture(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[0][1] );    
    glActiveTexture(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[0][0] );

    glBegin(GL_QUADS);
    glTexCoord2f(0,0);   glVertex2f(0,height);
    glTexCoord2f(0,mt);  glVertex2f(0,0);
    glTexCoord2f(ms,mt); glVertex2f(width,0);
    glTexCoord2f(ms,0);  glVertex2f(width,height);
    glEnd();

    glActiveTexture(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[1][1] );    
    glActiveTexture(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[1][0] );

    glBegin(GL_QUADS);
    glTexCoord2f(0,0);   glVertex2f(width,height);
    glTexCoord2f(0,mt);  glVertex2f(width,0);
    glTexCoord2f(ms,mt); glVertex2f(width * 2,0);
    glTexCoord2f(ms,0);  glVertex2f(width * 2,height);
    glEnd();

    glUseProgramObjectARB(0);

    glActiveTexture(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
    glActiveTexture(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);

    // Draw a cross hair
    glActiveTexture(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
    glActiveTexture(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
    
    glLineWidth(3);
    glColor3f(0, .8, 0);

    glBegin(GL_LINE_STRIP);
    {
        glVertex2f( width / 2 - 12, height / 2 - 12);
        glVertex2f( width / 2 - 12, height / 2 + 12);
        glVertex2f( width / 2 + 12, height / 2 + 12);
        glVertex2f( width / 2 + 12, height / 2 - 12);
        glVertex2f( width / 2 - 12, height / 2 - 12);
    }
    glEnd();

    glBegin(GL_LINE_STRIP);
    {
        glVertex2f( width / 2 - 12 + width, height / 2 - 12);
        glVertex2f( width / 2 - 12 + width, height / 2 + 12);
        glVertex2f( width / 2 + 12 + width, height / 2 + 12);
        glVertex2f( width / 2 + 12 + width, height / 2 - 12);
        glVertex2f( width / 2 - 12 + width, height / 2 - 12);
    }
    glEnd();

    glColor3f(1,1,1);
    glLineWidth(1);

    // IF THE USER WANTS DRAW THE SIDE VIEWS
   
    if( DrawSceneSampleView )
    {
        glBindTexture(GL_TEXTURE_2D, SceneSampleColorTex );

        glBegin(GL_QUADS);
        glTexCoord2f(0,   0);    glVertex2f(20,180);
        glTexCoord2f(0,   1.0);  glVertex2f(20,20);
        glTexCoord2f(1.0, 1.0);  glVertex2f(240,20);
        glTexCoord2f(1.0, 0);    glVertex2f(240,180);
        glEnd();
    }

    if( DrawPassTwoViews )
    {
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[0][1] );
        
        glBegin(GL_QUADS);
        glTexCoord2f(0,  0);    glVertex2f(20,380);
        glTexCoord2f(0,  mt);   glVertex2f(20,220);
        glTexCoord2f(ms, mt);   glVertex2f(240,220);
        glTexCoord2f(ms, 0);    glVertex2f(240,380);
        glEnd();

        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[1][1] );
        
        glBegin(GL_QUADS);
        glTexCoord2f(0,  0);    glVertex2f(20,580);
        glTexCoord2f(0,  mt);   glVertex2f(20,420);
        glTexCoord2f(ms, mt);   glVertex2f(240,420);
        glTexCoord2f(ms, 0);    glVertex2f(240,580);
        glEnd();
    }

    if( Capture )
    {
        snap("dfb_geowall.png",width * 2,height,FRAME_SRC);
    }

}

//========================================================================
// Drawing to the on-screen framebuffer in Personal Varrier Mode using
// Bob's varrier_combiner code
//========================================================================
void DrawToDefaultFB_Varrier()
{
    float ms, mt;
    ms = float(width);
    mt = float(height);

    glUseProgramObjectARB(PassThreePrg);

    for( int i = 0; i < Eyes; i++)
    {
        vc_prepare(&VCDisp,i);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluOrtho2D(0,width,height,0);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        glDisable(GL_LIGHTING);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_CULL_FACE);

        glActiveTexture(GL_TEXTURE1_ARB);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[i][1] );    
        glActiveTexture(GL_TEXTURE0_ARB);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[i][0] );

        glBegin(GL_QUADS);
        glTexCoord2f(0,0);   glVertex2f(0,height);
        glTexCoord2f(0,mt);  glVertex2f(0,0);
        glTexCoord2f(ms,mt); glVertex2f(width,0);
        glTexCoord2f(ms,0);  glVertex2f(width,height);
        glEnd();

        // Draw a cross hair
        glActiveTexture(GL_TEXTURE1_ARB);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
        glActiveTexture(GL_TEXTURE0_ARB);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);

        glLineWidth(3);
        glColor3f(0, .8, 0);
        glBegin(GL_LINE_STRIP);
        {
            glVertex2f( width / 2 - 12, height / 2 - 12);
            glVertex2f( width / 2 - 12, height / 2 + 12);
            glVertex2f( width / 2 + 12, height / 2 + 12);
            glVertex2f( width / 2 + 12, height / 2 - 12);
            glVertex2f( width / 2 - 12, height / 2 - 12);
        }
        glEnd();
        glLineWidth(1);
        glColor3f(1,1,1);
    
    }
    
    vc_combine( &VCDisp, LEP, REP );
}

//========================================================================
// Draw driver function
//========================================================================
void DrawScene()
{
    if( DisplayMode == PERSONAL_VARRIER_MODE )
    {
        if( tracker_position(0,LEP) )
        {
            LEP[0] -= .01f;
            LEP[1] -= 4.0f;
        }

        if( tracker_position(0,REP) )
        {
            REP[0] += .1f;
            REP[1] -= 4.0f;
        }

        printf("Right eye: %f , %f, %f\n", REP[0], REP[1], REP[2]);
        printf("Left  eye: %f , %f, %f\n", LEP[0], LEP[1], LEP[2]);
 //       LEP[0] = -.1;
  //      REP[0] = .1; 
   }

    float pos[3];
    float deltaY;



    switch(DisplayMode)
    {
    case MONO_MODE:
        DrawToSceneSampleTex();
        DOFCoefficients[FOCAL_DISTANCE] = RunSceneSampleFilter();
        DrawToColorBlurTex_Mono();
        DrawToDefaultFB_Mono();
        break;
    case GEOWALL_STEREO_MODE:
        get_camera_position( pos, pos + 1, pos + 2);
        position_camera( pos[0], pos[1], pos[2]);
        {
            DrawToSceneSampleTex();
            DOFCoefficients[FOCAL_DISTANCE] = RunSceneSampleFilter();
            DrawToColorBlurTex_Stereo();
            DrawToDefaultFB_Geowall();
        }
        position_camera( pos[0], pos[1], pos[2]);
        break;
    case PERSONAL_VARRIER_MODE:
        get_camera_position( pos, pos + 1, pos + 2);
        position_camera( pos[0], pos[1] - LEP[1], pos[2]);
        {
            DrawToSceneSampleTex();
            DOFCoefficients[FOCAL_DISTANCE] = RunSceneSampleFilter();
            DrawToColorBlurTex_Stereo();
            DrawToDefaultFB_Varrier();
        }
        position_camera(pos[0], pos[1], pos[2]);
        break;
    }

    if( Capture )
        Capture = false;

    SDL_GL_SwapBuffers();
}


//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void ProcessMouseMotion(SDL_Event* e)
{
    CamPitch += (e->motion.yrel) / float(height) * 180.0f;
    CamYaw   -= (e->motion.xrel) / float(width ) * 360.0f;
    orient_camera(DEG2RAD(CamPitch), DEG2RAD(CamYaw), 0.0f);
//    orient_model( ToriModel, DEG2RAD(CamPitch), DEG2RAD(CamYaw), 0.0f);
}

//----------------------------------------------------------------
// Process key up presses
// 
//----------------------------------------------------------------
void ProcessKeyUp(SDL_Event* e)
{
    switch( e->key.keysym.sym )
    {
    case SDLK_LEFT:
    case SDLK_RIGHT:
    case SDLK_UP:
    case SDLK_DOWN:
        LightVel[0] = LightVel[1] = LightVel[2] = 0;
        break;
    }

    if( (e->key.keysym.unicode & 0xFF80) == 0)
    {
        // it's an ascii character
        char ch = e->key.keysym.unicode & 0x7F;
        switch( ch )
        {
        case 'w':
            CamVel[2] = 0;
            break;
        case 's':
            CamVel[2] = 0;
            break;
        case 'a':
            CamVel[0] = 0;
            break;
        case 'd':
            CamVel[0] = 0;
            break;
        case 'k':
        case 'i':
            DOFDelta[FOCAL_DISTANCE] = 0.0;
            break;
        case 'j':
        case 'l':
            DOFDelta[FOCAL_RANGE] = 0.0;
            break;
        case '[':
        case ']':
            DOFDelta[MAX_COC] = 0.0;
            break;
        }
    }
}

//----------------------------------------------------------------
// Process key down presses
// 
//----------------------------------------------------------------
void ProcessKeyDown(SDL_Event* e)
{
    switch( e->key.keysym.sym )
    {
    case SDLK_ESCAPE:
        BreakLoop = true;
        break;
    case SDLK_LEFT:
        LightVel[0] = -10;
        break;
    case SDLK_RIGHT:
        LightVel[0] = 10;
        break;
    case SDLK_UP:
        LightVel[2] = 10;
        break;
    case SDLK_DOWN:
        LightVel[2] = -10;
        break;
    }


    if( (e->key.keysym.unicode & 0xFF80) == 0)
    {
        // it's an ascii character
        char ch = e->key.keysym.unicode & 0x7F;
        switch( ch )
        {
        case 'w':
            CamVel[2] = 10;
            break;
        case 's':
            CamVel[2] = -10;
            break;
        case 'a':
            CamVel[0] = 10;
            break;
        case 'd':
            CamVel[0] = -10;
            break;
        case 'c':
        case 'C':
            Capture = true;
            break;
        case 'x':
        case 'X':

            bind_model_shader_prg(RoomModel);
            if( !DisplayDepth )
            {
                glUniform1i(get_model_uniform_loc(RoomModel,"show_depth"), 1);
                DisplayDepth = true;
            }
            else
            {
                glUniform1i(get_model_uniform_loc(RoomModel,"show_depth"), 0);
                DisplayDepth = false;
            }
            printf("Toggling displaying depth buffer!\n");
            break;
        case 'i':
//            DOFDelta[FOCAL_DISTANCE] += .2;
            break;
        case 'k':
//            DOFDelta[FOCAL_DISTANCE] -= .2;
            break;
        case 'j':
            DOFDelta[FOCAL_RANGE] -= .2;
            break;
        case 'l':
            DOFDelta[FOCAL_RANGE] += .2;
            break;
        case '[':
            DOFDelta[MAX_COC] -= .2;
            break;
        case ']':
            DOFDelta[MAX_COC] += .2;
            break;
        case '<':
            FOV *= .5;
            break;
        case '>':
            FOV *= 2.0;
            break;
        case '1':
            DrawSceneSampleView = DrawSceneSampleView ? false : true;
            break;
        case '2':
            DrawPassTwoViews = DrawPassTwoViews ? false : true;
            break;
        case ';':
            VCDisp.shift -= .00001f;
            break;
        case '\'':
            VCDisp.shift += .00001f;
            break;
        case '-':
            LEP[1] -= .2;
            REP[1] -= .2;
            printf("Height %f\n", LEP[1]);
            break;
        case '=':
            LEP[1] += .2;
            REP[1] += .2;
            printf("Height %f\n", LEP[1]);
            break;
        }

    }
}


//----------------------------------------------------------------
// Infinite Loop
// 
//----------------------------------------------------------------
void InfLoop()
{
    double starttime, endtime;
    starttime = clock();
    endtime = starttime;
    int dt = 0;
    starttime = (double) SDL_GetTicks() / 1000.0; 
    
    //--- infinite loop with event queue processing
    SDL_Event event;
    while(!BreakLoop) 
    {
        while( SDL_PollEvent( &event )) 
        {
            switch( event.type )
            {
            case SDL_KEYDOWN:
                ProcessKeyDown(&event);
                break;
            case SDL_KEYUP:
                ProcessKeyUp(&event);
                break;
            case SDL_MOUSEMOTION:
                ProcessMouseMotion(&event);
                break;
            case SDL_QUIT:
                BreakLoop = true;
                break;
            }

        } // -- while event in queue

       
        //capture frame time
        endtime = (double) SDL_GetTicks() / 1000.0; 
        
        if( endtime - starttime > 1.0f / float(DESIRED_FPS) )
        {
            Update(endtime - starttime);
            starttime = (double) SDL_GetTicks() / 1000.0;
            
            // update scene 
            DrawScene();
        }

        
    } // -- infinite loop
}


//----------------------------------------------------------------
// homebrew update funciton
// 
//----------------------------------------------------------------
void Update(double dt)
{

/*    for( int i = 0; i < 3; i++)
    {
        LightPos[i] += LightVel[i] * dt;
    }
*/
    translate_camera( CamVel[0] * dt, CamVel[1] * dt, CamVel[2] * dt);

    bool print_update = false;
    for( int i = 0; i < 6; i++)
    {
        if( DOFDelta[i] != 0.0f)
            print_update = true;
        DOFCoefficients[i] += dt * DOFDelta[i];
    }

    if( print_update )
    {
        glUseProgramObjectARB(PassTwoPrg);
        glUniform1f(glGetUniformLocationARB(PassTwoPrg,"focal_dist"),
                    DOFCoefficients[FOCAL_DISTANCE]);
        glUniform1f(glGetUniformLocationARB(PassTwoPrg,"focal_range"),
                    DOFCoefficients[FOCAL_RANGE]);

        glUseProgramObjectARB(PassThreePrg);
        glUniform1f(glGetUniformLocationARB(PassThreePrg,"max_coc"),
                    DOFCoefficients[MAX_COC]);

        printf("!!!!------------DOF SETTINGS---------------------!!!!\n");
        printf("focal_dist %f\tdlens %f\nzfocus %f\tmax_coc %f\n",
               DOFCoefficients[FOCAL_DISTANCE], DOFCoefficients[LENS_DIAMETER],
               DOFCoefficients[ZFOCUS], DOFCoefficients[MAX_COC]);
        printf("focal_range %f\tscale %f\n", 
               DOFCoefficients[FOCAL_RANGE], DOFCoefficients[SCENE_RANGE]);

    }

    TotalTime += dt;
}


//----------------------------------------------------------------
// Setup lights
// 
//----------------------------------------------------------------
void InitLight()
{
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    
    GLfloat lightamb[] = { .3, .3, .3, 1 };
    GLfloat lightdif[] = { 1,  1,  1,  1 };
    glShadeModel(GL_SMOOTH);
    glLightfv(GL_LIGHT0, GL_AMBIENT,  lightamb);
    glLightfv(GL_LIGHT0, GL_DIFFUSE,  lightdif);
    glLightfv(GL_LIGHT0, GL_POSITION, LightPos);

}


//----------------------------------------------------------------
// SETUP SDL
// 
//----------------------------------------------------------------
void SetupSDL()
{
    // init video system
    const SDL_VideoInfo* vidinfo;
    if( SDL_Init(SDL_INIT_VIDEO) < 0 )
    {
        fprintf(stderr,"Failed to initialize SDL Video!\n");
        exit(1);
    }
    
    // tell system which funciton to process when exit() call is made
    atexit(SDL_Quit);

    SDL_EnableUNICODE(1);

    // get optimal video settings
    vidinfo = SDL_GetVideoInfo();
    if(!vidinfo)
    {
        fprintf(stderr,"Coudn't get video information!\n%s\n", SDL_GetError());
        exit(1);
    }

    // set opengl attributes
    SDL_GL_SetAttribute(SDL_GL_RED_SIZE,        5);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,      5);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,       5);
    SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE,      5);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,      32);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,    1);

    // get a framebuffer
    if( DisplayMode != GEOWALL_STEREO_MODE )
    {
        gDrawSurface = SDL_SetVideoMode(width,height,
                                        vidinfo->vfmt->BitsPerPixel,
                                        SDL_OPENGL | SDL_NOFRAME);
    }
    else
    {
        gDrawSurface = SDL_SetVideoMode(width * 2, height,
                                        vidinfo->vfmt->BitsPerPixel,
                                        SDL_OPENGL | SDL_NOFRAME);
        
    }

    if( !gDrawSurface )
    {
        fprintf(stderr,"Couldn't set video mode!\n%s\n", SDL_GetError());
        fprintf(stderr,"Trying lower depth size to 24\n");
        SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,      24);

        if( DisplayMode != GEOWALL_STEREO_MODE)
        {
            gDrawSurface = SDL_SetVideoMode(width,height,
                                            vidinfo->vfmt->BitsPerPixel, 
                                            SDL_OPENGL | SDL_NOFRAME);
        }
        else
        {
            gDrawSurface = SDL_SetVideoMode(width * 2, height,
                                            vidinfo->vfmt->BitsPerPixel,
                                            SDL_OPENGL | SDL_NOFRAME);
        }

        if(!gDrawSurface)
        {
            fprintf(stderr,"Couldn't set video mode!\n%s\n", SDL_GetError());
            fprintf(stderr,"Trying lower depth size 16\n");
            SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,      16);

            if( DisplayMode != GEOWALL_STEREO_MODE)
            {
                gDrawSurface = SDL_SetVideoMode(width,height,
                                                vidinfo->vfmt->BitsPerPixel, 
                                                SDL_OPENGL | SDL_NOFRAME);
            }
            else
            {
                gDrawSurface = SDL_SetVideoMode(width * 2, height,
                                                vidinfo->vfmt->BitsPerPixel,
                                                SDL_OPENGL | SDL_NOFRAME);
            }

            if(!gDrawSurface)
            {
                fprintf(stderr,"Couldn't set video mode!\n%s\n", 
                    SDL_GetError());
                exit(1);
            }
        }
    }

    // give the window a name
    SDL_WM_SetCaption("ARUN RAO - DOF GLSL", "DOF GLSL");
    SDL_ShowCursor(0);
    SDL_WM_GrabInput(SDL_GRAB_ON);

    // set opengl viewport and perspective view
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glViewport(0,0,width,height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective( 60, 4.0f / 3.0f, .01, 10000);
    glMatrixMode(GL_MODELVIEW);
}

#ifdef __APPLE__
//----------------------------------------------------------------
void SetupGLProcs()
{
}

#elif !defined(_WIN32)

//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void SetupGLProcs()
{
    // SHADER PROCS
    glCreateProgramObjectARB = (PFNGLCREATEPROGRAMOBJECTARBPROC)
        glGetProcAddress("glCreateProgramObjectARB");
    glCreateShaderObjectARB = (PFNGLCREATESHADEROBJECTARBPROC)
        glGetProcAddress("glCreateShaderObjectARB");
    glShaderSourceARB = (PFNGLSHADERSOURCEARBPROC)
        glGetProcAddress("glShaderSourceARB");
    glCompileShaderARB = (PFNGLCOMPILESHADERARBPROC)
        glGetProcAddress("glCompileShaderARB");
    glGetObjectParameterivARB = (PFNGLGETOBJECTPARAMETERIVARBPROC)
        glGetProcAddress("glGetObjectParameterivARB");
    glAttachObjectARB = (PFNGLATTACHOBJECTARBPROC)
        glGetProcAddress("glAttachObjectARB");
    glGetInfoLogARB = (PFNGLGETINFOLOGARBPROC)
        glGetProcAddress("glGetInfoLogARB");
    glLinkProgramARB = (PFNGLLINKPROGRAMARBPROC)
        glGetProcAddress("glLinkProgramARB");
    glUseProgramObjectARB = (PFNGLUSEPROGRAMOBJECTARBPROC)
        glGetProcAddress("glUseProgramObjectARB");
    glGetUniformLocationARB = (PFNGLGETUNIFORMLOCATIONARBPROC)
        glGetProcAddress("glGetUniformLocationARB");
    glUniform1f = (PFNGLUNIFORM1FARBPROC)
        glGetProcAddress("glUniform1fARB");
    glUniform2f = (PFNGLUNIFORM2FARBPROC)
        glGetProcAddress("glUniform2fARB");
    glUniform3f = (PFNGLUNIFORM3FARBPROC)
        glGetProcAddress("glUniform3fARB"); 
    glUniform1fv = (PFNGLUNIFORM1FVARBPROC)
        glGetProcAddress("glUniform1fvARB");
    glUniform2fv = (PFNGLUNIFORM2FVARBPROC)
        glGetProcAddress("glUniform2fvARB");
    glUniform3fv = (PFNGLUNIFORM3FVARBPROC)
        glGetProcAddress("glUniform3fvARB"); 
    glUniform1i = (PFNGLUNIFORM1IARBPROC)
        glGetProcAddress("glUniform1iARB");
    glDeleteObjectARB = (PFNGLDELETEOBJECTARBPROC)
        glGetProcAddress("glDeleteObjectARB");
    glBindAttribLocationARB = (PFNGLBINDATTRIBLOCATIONARBPROC)
        glGetProcAddress("glBindAttribLocationARB");
    glGetAttribLocationARB = (PFNGLGETATTRIBLOCATIONARBPROC)
        glGetProcAddress("glGetAttribLocationARB");
    glVertexAttrib3fARB = (PFNGLVERTEXATTRIB3FARBPROC)
        glGetProcAddress("glVertexAttrib3fARB");
    

    // FBO FUNCTIONS
    glDrawBuffers = (PFNGLDRAWBUFFERSPROC)
        glGetProcAddress("glDrawBuffers");
    glGenFramebuffersEXT = (PFNGLGENFRAMEBUFFERSEXTPROC)
        glGetProcAddress("glGenFramebuffersEXT");
    glBindFramebufferEXT = (PFNGLBINDFRAMEBUFFEREXTPROC)
        glGetProcAddress("glBindFramebufferEXT");
    glDeleteFramebuffersEXT = (PFNGLDELETEFRAMEBUFFERSEXTPROC)
        glGetProcAddress("glDeleteFramebuffersEXT");
    glFramebufferTexture2DEXT = (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC)
        glGetProcAddress("glFramebufferTexture2DEXT");
    glCheckFramebufferStatusEXT = (PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC)
        glGetProcAddress("glCheckFramebufferStatusEXT");

    // MULTITEXTURE FUNCTIONS
    glActiveTexture = (PFNGLACTIVETEXTUREPROC)
        glGetProcAddress("glActiveTexture");
}

#endif


//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void CheckGLErrors(const char* message)
{
    printf("%s ", message);
    switch( glGetError() )
    {
    case GL_NO_ERROR: break;
    case GL_INVALID_ENUM: 
        printf("GL ERR: Unacceptable value for enum argument");
        break;
    case GL_INVALID_VALUE: 
        printf("GL ERR: Numeric argument out of range");
        break;
    case GL_INVALID_OPERATION:
        printf("GL ERR: Operation not allowed in current state");
        break;
    case GL_STACK_OVERFLOW:
        printf("GL ERR: Command would cause stack overflow");
        break;
    case GL_STACK_UNDERFLOW:
        printf("GL ERR: Command would cause stack underflow");
        break;
    case GL_OUT_OF_MEMORY:
        printf("GL ERR: Not enough memory left to execute command");
        break;
    case GL_TABLE_TOO_LARGE:
        printf("GL ERR: Specified table exceeds implementations max");
        break;
    }
    printf("\n");
}

//----------------------------------------------------------------
// 
//----------------------------------------------------------------
void PrintFBOStatus(const char* message)
{

    printf("%s FBO Status: ", message);
    switch( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT))
    {
    case GL_FRAMEBUFFER_COMPLETE_EXT: 
        printf("Complete!"); 
        break;
    case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
        printf("Incomplete Attachment.");
        break;
    case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
        printf("Incomplte Missing Attachment.");
        break;
    case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
        printf("Incomplete Problems with Dimensions.");
        break;
    case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
        printf("Incomplete Problems with Attachment Formats.");
        break;
    case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
        printf("Incomplete Problems with Draw Buffer.");
        break;
    case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
        printf("Incomplete Problems with Read Buffer.");
        break;
    case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
        printf("Unsupported.  Check GL version!");
        break;
    }
    printf("\n");
}

//=========================================================================
void LoadModels()
{

    printf("\n---\n");
    RoomModel = load_model("models/room.obj", NULL, NULL);
    scale_model(RoomModel, 2.5, 1.0, 1.0);
    printf("\n---\n");
    ToriModel = load_model("models/tori.obj", RoomModel);
    position_model(ToriModel, 4, -3.0, 0);
    printf("\n---\n");
    GunModel  = load_model("models/gun.obj", RoomModel);

    
}

void SetupFBO()
{
    int i;
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_TEXTURE_RECTANGLE_ARB);
 
    // Color Blur FBO per Eye!

    ColorBlurFBO      = new GLuint[Eyes];
    ColorBlurColorTex = new GLuint*[Eyes];
    ColorBlurDepthTex = new GLuint[Eyes];

    glGenFramebuffersEXT( Eyes, ColorBlurFBO );
    glGenTextures( Eyes, ColorBlurDepthTex );
    
    for( i = 0; i < Eyes; i++)
    {
        // Build two color attachments and one depth per eye
        
        ColorBlurColorTex[i] = new GLuint[2];
        glGenTextures( 2, ColorBlurColorTex[i] );

        for( int k = 0; k < 2; k++)
        {
            glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurColorTex[i][k]);
            glTexImage2D(GL_TEXTURE_RECTANGLE_ARB,0,GL_RGBA16, width, height, 
                         0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, 
                            GL_CLAMP );
            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, 
                            GL_CLAMP);
            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, 
                            GL_LINEAR);
            glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, 
                            GL_LINEAR);
        }

        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ColorBlurDepthTex[i]);
        glTexImage2D(GL_TEXTURE_RECTANGLE_ARB,0,GL_DEPTH_COMPONENT, 
                     width, height, 0, GL_DEPTH_COMPONENT, 
                     GL_UNSIGNED_BYTE, NULL);
        glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S,GL_CLAMP );
        glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T,GL_CLAMP );
        glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, 
                        GL_LINEAR);
        glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, 
                        GL_LINEAR);

        // Bind and Attach
        
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, ColorBlurFBO[i] );
        glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT,
                                   GL_DEPTH_ATTACHMENT_EXT,
                                   GL_TEXTURE_RECTANGLE_ARB, 
                                   ColorBlurDepthTex[i], 
                                   0);

        glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT,
                                   GL_COLOR_ATTACHMENT0_EXT,
                                   GL_TEXTURE_RECTANGLE_ARB, 
                                   ColorBlurColorTex[i][0],  
                                   0);


        glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT,
                                   GL_COLOR_ATTACHMENT1_EXT,
                                   GL_TEXTURE_RECTANGLE_ARB, 
                                   ColorBlurColorTex[i][1],  
                                   0);

        printf("Color Blur %d ", i);
        PrintFBOStatus("Checking");
    }


    // Scene Sample FBO
    
    glGenFramebuffersEXT(1, &SceneSampleFBO );
    glGenTextures( 1, &SceneSampleColorTex );
    glGenTextures( 1, &SceneSampleDepthTex );

    glBindTexture(GL_TEXTURE_2D, SceneSampleColorTex);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16, SCENE_SAMPLE_DIM, 
                 SCENE_SAMPLE_DIM, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        
    glBindTexture(GL_TEXTURE_2D, SceneSampleDepthTex);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, SCENE_SAMPLE_DIM,
                 SCENE_SAMPLE_DIM, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, 
                 NULL);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, SceneSampleFBO );
    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT,
                               GL_DEPTH_ATTACHMENT_EXT,
                               GL_TEXTURE_2D,
                               SceneSampleDepthTex,
                               0);
    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT,
                               GL_COLOR_ATTACHMENT0_EXT,
                               GL_TEXTURE_2D,
                               SceneSampleColorTex,
                               0);

    PrintFBOStatus("Checking Scene Sample");

}



GLfloat SampleFilter[ SCENE_SAMPLE_DIM ];
GLfloat DepthSample[SCENE_SAMPLE_DIM * SCENE_SAMPLE_DIM];
GLfloat rows[SCENE_SAMPLE_DIM];

void SetupSceneSampleFilter()
{
    int mid = SCENE_SAMPLE_DIM / 2;
    int i;
    float sum = 0.0f;
    printf("MID: %d\n", mid);
    for( i = mid - 1; i >= 0; --i)
    {
        SampleFilter[i] = ((float) i) / ((float) (SCENE_SAMPLE_DIM / 2) + 1.0);
        SampleFilter[i] = SampleFilter[i] * SampleFilter[i];
        SampleFilter[31 - i] = SampleFilter[i];
        sum += 2 * SampleFilter[i];        
    }

    for(i = 0; i < SCENE_SAMPLE_DIM; i++)
    {
        SampleFilter[i] /= sum;
        printf("Filter Element %d: %f\n", i, SampleFilter[i]);
    }

    sum = 0.0f;

    for( i = 0; i < SCENE_SAMPLE_DIM; i++)
        sum += SampleFilter[i];

    printf("Sum of Weights: %f\n", sum);
}


float RunSceneSampleFilter()
{
    int i;

    glBindTexture(GL_TEXTURE_2D, SceneSampleColorTex );
    glGetTexImage(GL_TEXTURE_2D, 0, GL_RED, GL_FLOAT, DepthSample);

    for( i = 0; i < SCENE_SAMPLE_DIM; ++i)
    {
        rows[i] = 0.0f;
        for(int k = 0; k < SCENE_SAMPLE_DIM; k++)
        {
            rows[i] += (DepthSample[i * SCENE_SAMPLE_DIM + k] * 
                        (FAR_PLANE - NEAR_PLANE) + NEAR_PLANE) * 
                SampleFilter[k];
        }
    }
    
    float value = 0.0f;
    for(i = 0; i < SCENE_SAMPLE_DIM; ++i)
        value += rows[i] * SampleFilter[i];

    return value;

}

void SetupMultiplePassShaders()
{
    SetupShaders("shaders/passone.vert","shaders/passone.frag",
                PassOnePrg, PassOneVs, PassOneFs);
    glUniform1f(glGetUniformLocationARB(PassOnePrg,"nplane"), NEAR_PLANE);
    glUniform1f(glGetUniformLocationARB(PassOnePrg,"fplane"), FAR_PLANE);

    SetupShaders("shaders/passtwo.vert","shaders/passtwo.frag",
                 PassTwoPrg, PassTwoVs, PassTwoFs);
    glUniform1i(glGetUniformLocationARB(PassTwoPrg,"cmap"),    0);


    float FocalDistance = 0.5f;
    float LensDiameter = 2.0f;
    float ZFocus = 5.0f;
    float MaxCOC = 12.0f;
    float SceneRange = FAR_PLANE - NEAR_PLANE;
    float FocalRange = 6.0f;

    for( int i = 0; i < 6; i++)
        DOFDelta[i] = 0.0;

    DOFCoefficients[FOCAL_DISTANCE] = FocalDistance;
    DOFCoefficients[LENS_DIAMETER]  = LensDiameter;
    DOFCoefficients[ZFOCUS]         = ZFocus;
    DOFCoefficients[MAX_COC]        = MaxCOC;
    DOFCoefficients[SCENE_RANGE]    = SceneRange;
    DOFCoefficients[FOCAL_RANGE]    = FocalRange;

    glUniform1f(glGetUniformLocationARB(PassTwoPrg,"focal_dist"),
                FocalDistance);
    glUniform1f(glGetUniformLocationARB(PassTwoPrg,"focal_range"), 
                FocalRange);
    glUniform1f(glGetUniformLocationARB(PassTwoPrg,"scene_range"),
                SceneRange);

    SetupShaders("shaders/passthree.vert","shaders/passthree.frag",
                 PassThreePrg, PassThreeVs, PassThreeFs);
    glUniform1i(glGetUniformLocationARB(PassThreePrg,"cmap"),  0);
    glUniform1i(glGetUniformLocationARB(PassThreePrg,"dbmap"), 1);
    glUniform1f(glGetUniformLocationARB(PassThreePrg,"max_coc"),
                DOFCoefficients[MAX_COC]);

    // build pass three filter kernel
    GLfloat u[12], v[12];
    GLfloat sx, sy;
    sx = 1.0;
    sy = 1.0;

    u[0]  = -0.326212f;     v[0]  = -0.40581f;
    u[1]  = -0.840144f;     v[1]  = -0.07358f;
    u[2]  = -0.695914f;     v[2]  = 0.457137f;
    u[3]  = -0.203345f;     v[3]  = 0.620716f;
    u[4]  = 0.96234f;       v[4]  = -0.194983f;
    u[5]  = 0.473434f;      v[5]  = -0.480026f;
    u[6]  = 0.519456f;      v[6]  = 0.767022f;
    u[7]  = 0.185461f;      v[7]  = -0.893124f;
    u[8]  = 0.507431f;      v[8]  = 0.064425f;
    u[9]  = 0.89642f;       v[9]  = 0.412458f;
    u[10] = -0.32194f;      v[10] = -0.932615f;
    u[11] = -0.791559f;     v[11] = -0.59771f;

    for( int i = 0; i < 12; ++i)
    {
        u[i] *= sx;
        v[i] *= sy;
        char uniform_name[16];
        char index[3];
        sprintf(index,"%d",i);
        memset(uniform_name, 0, 16);
        strcat(uniform_name,"tap_array[");
        strcat(uniform_name,index);
        strcat(uniform_name,"]");
        printf("Assigning uniform %s value of %f, %f\n", uniform_name,
               u[i],v[i]);
        glUniform2f(glGetUniformLocationARB(PassThreePrg,uniform_name),
                    u[i], v[i]);
    }
    
}
